/* 
    Webcam Lab

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    (C) 2005 Jason Hunt
    nulluser@gmail.com
    
    file: capture.cpp
*/

/* 
    Code copied and adapted from  Limitz - http://paste.rohitab.com/560
*/


#include "core.h"



//VIDEODEVICE
//public
VideoDevice::VideoDevice()
{
	friendlyname = (char*) calloc(1, MAX_DEVICE_NAME * sizeof(char));
	filtername =  (WCHAR*) calloc(1, MAX_DEVICE_NAME * sizeof(WCHAR));
	filter = 0;
}

VideoDevice::~VideoDevice()
{
	free(friendlyname);
	free(filtername);
}

const char* VideoDevice::GetFriendlyName()
{
	return friendlyname;
}


//VIDEOCAPTURE
//public
VideoCapture::VideoCapture(VideoCaptureCallback cb)
{
    ready = 0;

	CoInitialize(NULL);

	playing = false;
	current = 0;
	callbackhandler = new CallbackHandler(cb);
	devices = new VideoDevice[MAX_DEVICES];

	InitializeGraph();
	SetSourceFilters();
	SetSampleGrabber();
	SetNullRenderer();
	
//	    SetResolution( 640, 480);
	    
}

VideoCapture::~VideoCapture()
{
	delete callbackhandler;
	delete[] devices;
}

VideoDevice* VideoCapture::GetDevices()
{
	return devices;
}

int VideoCapture::NumDevices()
{
	return num_devices;
}

void VideoCapture::Select(VideoDevice* dev)
{
	HRESULT hr;
	LONGLONG start=MAXLONG, stop=MAXLONG;
	bool was_playing = playing;

	if (!dev->filter) throw E_INVALIDARG;

	//temporary stop
	if (playing) Stop();

	if (current)
	{
		//remove and add the filters to force disconnect of pins
		graph->RemoveFilter(current->filter);
		graph->RemoveFilter(samplegrabberfilter);

		graph->AddFilter(samplegrabberfilter,L"Sample Grabber");
		graph->AddFilter(current->filter, current->filtername);
	}
	
	start = 0;
	current = dev;

//SetRes();

    // setup special functions

   SetResolution(640,480);

    //SetExt();


	//connect graph with current source filter
	#ifdef SHOW_DEBUG_RENDERER
	hr = capture->RenderStream(&PIN_CATEGORY_CAPTURE, &MEDIATYPE_Video, current->filter, samplegrabberfilter, NULL);
	#else
	hr = capture->RenderStream(&PIN_CATEGORY_CAPTURE, &MEDIATYPE_Video, current->filter, samplegrabberfilter, nullrenderer);
	#endif
	if (hr != S_OK) throw hr;
		
	
	
    AM_MEDIA_TYPE media;
	

    samplegrabber->GetConnectedMediaType( &media ); 
 
    x_size = 0;
    y_size = 0;

	VIDEOINFOHEADER *pVideoHeader = (VIDEOINFOHEADER*)media.pbFormat;

    if (pVideoHeader)
	{	
	   x_size = pVideoHeader->bmiHeader.biWidth;
       y_size = pVideoHeader->bmiHeader.biHeight;
    }
    
	//start streaming
	hr = capture->ControlStream(&PIN_CATEGORY_CAPTURE, &MEDIATYPE_Video, current->filter, &start, &stop, 1,2);
	if (hr < 0) throw hr;

	//restart
	if (was_playing) Start();
	
	
}

void VideoCapture::Start()
{
	HRESULT hr;

	hr = control->Run();
	if (hr < 0) throw hr;

	playing = true;
	
	ready = 1;
	
}

void VideoCapture::Stop()
{
	HRESULT hr;

	hr = control->StopWhenReady();
	if (hr < 0) throw hr;

	playing = false;
	
	ready = 0;
}

//protected
void VideoCapture::InitializeGraph()
{
	HRESULT hr;

	//create the FilterGraph
	hr = CoCreateInstance(CLSID_FilterGraph,NULL,CLSCTX_INPROC_SERVER,IID_IFilterGraph2,(void**) &graph);
	if (hr < 0) throw hr; 
	
	//create the CaptureGraphBuilder
	hr = CoCreateInstance(CLSID_CaptureGraphBuilder2,NULL,CLSCTX_INPROC_SERVER,IID_ICaptureGraphBuilder2,(void**) &capture);
	if (hr < 0) throw hr;
	
	//get the controller for the graph
	hr = graph->QueryInterface(IID_IMediaControl, (void**) &control);
	if (hr < 0) throw hr;

	capture->SetFiltergraph(graph);
}
// Release the format block for a media type.

void _FreeMediaType(AM_MEDIA_TYPE& mt)
{
    if (mt.cbFormat != 0)
    {
        CoTaskMemFree((PVOID)mt.pbFormat);
        mt.cbFormat = 0;
        mt.pbFormat = NULL;
    }
    if (mt.pUnk != NULL)
    {
        // pUnk should not be used.
        mt.pUnk->Release();
        mt.pUnk = NULL;
    }
}
// Delete a media type structure that was allocated on the heap.
void _DeleteMediaType(AM_MEDIA_TYPE *pmt)
{
    if (pmt != NULL)
    {
        _FreeMediaType(*pmt); 
        CoTaskMemFree(pmt);
    }
}

void VideoCapture::SetResolution( int width, int height)
{
    IAMStreamConfig *config;
    config = NULL;
    
    int r = capture->FindInterface(&PIN_CATEGORY_CAPTURE, 
                           &MEDIATYPE_Video, 
                           current->filter, 
                           IID_IAMStreamConfig, 
                           (void**) &config);
                           
    if (r == E_FAIL)  { console("SetResolution( E_FAIL\n"); return; } else
    if (r == E_NOINTERFACE) { console("SetResolution( E_NOINTERFACE\n"); return; }else
    if (r == E_POINTER) { console("SetResolution( E_POINTER\n"); return; }


    int resolutions, size;
    
    VIDEO_STREAM_CONFIG_CAPS caps;
    
    config->GetNumberOfCapabilities(&resolutions, &size);

    for (int i = 0; i < resolutions; i++) 
    {
        AM_MEDIA_TYPE *mediaType;

        if (config->GetStreamCaps(i, &mediaType,   reinterpret_cast<BYTE*>(&caps)) == S_OK ) 
        {
            int maxWidth = caps.MaxOutputSize.cx;
            int maxHeight = caps.MaxOutputSize.cy;

            console("Max (%d, %d)\n", maxWidth, maxHeight);

            //console("Sub Type: %d\n", mediaType->subtype);
            //console("MEDIASUBTYPE_YUY2: %d\n",MEDIASUBTYPE_YUY2);


            if(maxWidth == width && maxHeight == height) 
            {

                mediaType->subtype = MEDIASUBTYPE_MJPG;




                VIDEOINFOHEADER *info =
                    reinterpret_cast<VIDEOINFOHEADER*>(mediaType->pbFormat);

                info->bmiHeader.biWidth = maxWidth; 
                info->bmiHeader.biHeight = maxHeight;
                info->bmiHeader.biSizeImage = DIBSIZE(info->bmiHeader); 
                config->SetFormat(mediaType);

                _DeleteMediaType(mediaType);
                break;

            }

            _DeleteMediaType(mediaType);
        }
    }
}


void VideoCapture::UpdateSize( void )
{
    AM_MEDIA_TYPE media;
	

    samplegrabber->GetConnectedMediaType( &media ); 
 
    x_size = 0;
    y_size = 0;

	VIDEOINFOHEADER *pVideoHeader = (VIDEOINFOHEADER*)media.pbFormat;

    if (pVideoHeader)
	{	
	   x_size = pVideoHeader->bmiHeader.biWidth;
       y_size = pVideoHeader->bmiHeader.biHeight;
    }
}




void VideoCapture::SetExt( void )
{

//IBaseFilter *pFilter;
/* Obtain the filter's IBaseFilter interface. (Not shown) */
ISpecifyPropertyPages *pProp;
HRESULT hr = current->filter->QueryInterface(IID_ISpecifyPropertyPages, (void **)&pProp);




if (SUCCEEDED(hr)) 
{
    // Get the filter's name and IUnknown pointer.
    FILTER_INFO FilterInfo;
    hr = current->filter->QueryFilterInfo(&FilterInfo); 
    IUnknown *pFilterUnk;
    current->filter->QueryInterface(IID_IUnknown, (void **)&pFilterUnk);

    // Show the page. 
    CAUUID caGUID;
    pProp->GetPages(&caGUID);
    pProp->Release();
   
    OleCreatePropertyFrame(
        NULL,                   // Parent window
        0, 0,                   // Reserved
        FilterInfo.achName,     // Caption for the dialog box
        1,                      // Number of objects (just the filter)
        &pFilterUnk,            // Array of object pointers. 
        caGUID.cElems,          // Number of property pages
        caGUID.pElems,          // Array of property page CLSIDs
        0,                      // Locale identifier
        0, NULL                 // Reserved
    );

    // Clean up.
    pFilterUnk->Release();
    FilterInfo.pGraph->Release(); 
    CoTaskMemFree(caGUID.pElems);
}
    
    
}

static void ShowPropertyPage( IUnknown *obj )
{
    ISpecifyPropertyPages *p_spec;
    CAUUID cauuid;

    HRESULT hr = obj->QueryInterface( IID_ISpecifyPropertyPages,
                                      (void **)&p_spec );
    if( FAILED(hr) ) return;

    if( SUCCEEDED(p_spec->GetPages( &cauuid )) )
    {
        if( cauuid.cElems > 0 )
        {
            HWND hwnd_desktop = ::GetDesktopWindow();

            OleCreatePropertyFrame( hwnd_desktop, 30, 30, NULL, 1, &obj,
                                    cauuid.cElems, cauuid.pElems, 0, 0, NULL );

            CoTaskMemFree( cauuid.pElems );
        }
        p_spec->Release();
    }
}


extern HWND system_hwnd;

void VideoCapture::SetRes( void )
{
   control->Stop(); // Stop the graph.
    

	IAMStreamConfig *prop;
	
	
	HRESULT hr = capture->FindInterface(&PIN_CATEGORY_CAPTURE,
                                        &MEDIATYPE_Video, current->filter,
                                        IID_IAMStreamConfig, (void **)&prop);
    ShowPropertyPage     (prop);                             
                       
                                      
   // UpdateSize();;
                                                     
                                                                                   
	
    control->Run();	
}




/*   
REF code
     
 IAMStreamConfig *pSC;
    hr = g_pCapture->FindInterface(&PIN_CATEGORY_CAPTURE,
        &MEDIATYPE_Video, pSrcFilter,
        IID_IAMStreamConfig, (void **)&pSC);

    ISpecifyPropertyPages *pSpec;
    CAUUID cauuid;

    hr = pSC->QueryInterface(IID_ISpecifyPropertyPages,
        (void **)&pSpec);

    if(hr == S_OK)
    {
        hr = pSpec->GetPages(&cauuid);
        hr = OleCreatePropertyFrame(ghApp, 30, 30, NULL, 1,
            (IUnknown **)&pSC, cauuid.cElems,
            (GUID *)cauuid.pElems, 0, 0, NULL);

        CoTaskMemFree(cauuid.pElems);
        pSpec->Release();
    }

    pSC->Release();
*/







void VideoCapture::SetSourceFilters()
{
	HRESULT hr;
	VARIANT name;
	LONGLONG start=MAXLONG, stop=MAXLONG;

	unsigned long dev_count;
	
	ICreateDevEnum*		dev_enum;
	IEnumMoniker*		enum_moniker;
	IMoniker*			moniker;
	IPropertyBag*		pbag;

	//create an enumerator for video input devices
	hr = CoCreateInstance(CLSID_SystemDeviceEnum, NULL,CLSCTX_INPROC_SERVER,IID_ICreateDevEnum,(void**) &dev_enum);
	if (hr < 0) throw hr;

	hr = dev_enum->CreateClassEnumerator(CLSID_VideoInputDeviceCategory,&enum_moniker,0);
	if (hr < 0) throw hr;
	if (hr == S_FALSE) return; //no devices found

	//get devices (max 8)
	num_devices = 0;
	enum_moniker->Next(MAX_DEVICES, &moniker, &dev_count);
	for (unsigned int i=0; i<dev_count; i++)
	{
		//get properties
		hr = moniker[i].BindToStorage(0, 0, IID_IPropertyBag, (void**) &pbag);
		if (hr >= 0)
		{
			VariantInit(&name);

			//get the description
			hr = pbag->Read(L"Description", &name, 0);
			if (hr < 0) hr = pbag->Read(L"FriendlyName", &name, 0);
			if (hr >= 0)
			{
				//Initialize the VideoDevice struct
				VideoDevice* dev = devices+num_devices++;
				BSTR ptr = name.bstrVal;

				for (int c = 0; *ptr; c++, ptr++)
				{
					//bit hacky, but i don't like to include ATL
					dev->filtername[c] = *ptr;
					dev->friendlyname[c] = *ptr & 0xFF;
					
			
					
				}

                console("fname: %s\n", dev->friendlyname );

				//add a filter for the device
				hr = graph->AddSourceFilterForMoniker(moniker+i, 0, dev->filtername, &dev->filter);
				if (hr != S_OK) num_devices--;
			}
			VariantClear(&name);
			pbag->Release();
		}
		moniker[i].Release();
	}
}

void VideoCapture::SetSampleGrabber()
{
	HRESULT hr;

	hr = CoCreateInstance(CLSID_SampleGrabber, NULL, CLSCTX_INPROC_SERVER,IID_IBaseFilter,(void**)&samplegrabberfilter);
	if (hr < 0) throw hr;

	hr = graph->AddFilter(samplegrabberfilter, L"Sample Grabber");
	if (hr != S_OK) throw hr;

	hr = samplegrabberfilter->QueryInterface(IID_ISampleGrabber, (void**)&samplegrabber);
	if (hr != S_OK) throw hr;

	//set the media type
	AM_MEDIA_TYPE mt;
	memset(&mt, 0, sizeof(AM_MEDIA_TYPE));

	mt.majortype	= MEDIATYPE_Video;
	mt.subtype		= MEDIASUBTYPE_RGB24; 
	
	// setting the above to 32 bits fails consecutive Select for some reason
	// and only sends one single callback (flush from previous one ???)
	// must be deeper problem. 24 bpp seems to work fine for now.
	
	callbackhandler->SetMediaType(&mt);
	hr = samplegrabber->SetMediaType(&mt);
	if (hr != S_OK) throw hr;


	samplegrabber->SetCallback(callbackhandler,0);

	
}

void VideoCapture::SetNullRenderer()
{
	HRESULT hr;

	hr = CoCreateInstance(CLSID_NullRenderer,NULL,CLSCTX_INPROC_SERVER,IID_IBaseFilter,(void**) &nullrenderer);
	if (hr < 0) throw hr; 

	graph->AddFilter(nullrenderer, L"Null Renderer");
}





/* Call back methods */





//CALLBACKHANDLER
//public
VideoCapture::CallbackHandler::CallbackHandler(VideoCaptureCallback cb)
{
	callback = cb;
}

VideoCapture::CallbackHandler::~CallbackHandler()
{
}

void VideoCapture::CallbackHandler::SetMediaType(AM_MEDIA_TYPE* am)
{
	if (am->subtype == MEDIASUBTYPE_RGB555) bitpixel = 16;
	else if (am->subtype == MEDIASUBTYPE_RGB24) bitpixel = 24;
	else if (am->subtype == MEDIASUBTYPE_RGB32) bitpixel = 32;
}

HRESULT VideoCapture::CallbackHandler::SampleCB(double time, IMediaSample *sample)
{
	HRESULT hr;
	AM_MEDIA_TYPE* mt;
	unsigned char* buffer;

	hr = sample->GetPointer((BYTE**)&buffer);
	if (hr != S_OK) return S_OK;

	hr = sample->GetMediaType(&mt);
	
	if (hr < 0) return S_OK;
	if (hr == S_OK) SetMediaType(mt);

	callback(buffer, sample->GetActualDataLength(), bitpixel);
	return S_OK;
}

HRESULT VideoCapture::CallbackHandler::BufferCB(double time, BYTE *buffer, long len)
{
	return S_OK;
}

HRESULT VideoCapture::CallbackHandler::QueryInterface(const IID &iid, LPVOID *ppv)
{
	if( iid == IID_ISampleGrabberCB || iid == IID_IUnknown )
	{
		*ppv = (void *) static_cast<ISampleGrabberCB*>( this );
		return S_OK;
	}
	return E_NOINTERFACE;
}

ULONG VideoCapture::CallbackHandler::AddRef()
{
	return 1;
}

ULONG VideoCapture::CallbackHandler::Release()
{
	return 2;
}




void VideoCapture::GetSize(int &xs, int &ys)
{
    xs = x_size;
    ys = y_size;
}





