/*
    Robot Interface
    (C) 2006 Jason Hunt
    nulluser@gmail.com
    
    File: config.cpp
*/

#include "core.h"


#define LINE_MAX 1024





/* gets the nextr string in a line, up to a space or endline */
void get_next_string(char *line, unsigned int &i, char *str)
{
    int done_str = 0;   // flag for done with string
    
    unsigned int s = 0; // index into target string

    // eat white space at front of line
    while ((line[i] == ' ' || line[i] == '\t') && i < LINE_MAX) i++;
  
    while (!done_str && i < LINE_MAX)
    {
        char ch = line[i++];
                               
        if (ch == ' ' || ch == '\t' ||  ch == 0)
            done_str = 1;
        else
            str[s++] = ch;
    }

    str[s] = 0;   // null terminate
}
/* end of get_next_string */


/* parse a config line for commands */
void parse_line(char *line)
{
    unsigned int index = 0;
    char str[LINE_MAX];
    
    get_next_string(line, index, str);
   
    if (!strcmp("speed_limit", str))                  
    {
        get_next_string(line, index, str);
    
        if (strlen(str))
        {
            speed_limit = atof(str);
            console("Set speed_limit %f\n", speed_limit);
        }
    } 


    if (!strcmp("min_time", str))                  
    {
        get_next_string(line, index, str);
    
        if (strlen(str))
        {
            min_time = atof(str);
            console("Set min_time %f\n", min_time);
        }
    } 

    if (!strcmp("dist", str))                  
    {
        get_next_string(line, index, str);
    
        if (strlen(str))
        {
            dist = atof(str);
            console("Set dist %f\n", dist);
        }
    } 


    if (!strcmp("timeout", str))                  
    {
        get_next_string(line, index, str);
    
        if (strlen(str))
        {
            timeout = atof(str);
            console("Set timeout %f\n", timeout);
        }
    } 



    if (!strcmp("clear", str))                  
    {
        get_next_string(line, index, str);
    
        if (strlen(str))
        {
            clear = atof(str);
            console("Set clear %f\n", clear);
        }
    } 


    if (!strcmp("alpha", str))                  
    {
        get_next_string(line, index, str);
    
        if (strlen(str))
        {
            alpha = atof(str);
            console("Set alpha %f\n", alpha);
        }
    } 



    if (!strcmp("act_alpha", str))                  
    {
        get_next_string(line, index, str);
    
        if (strlen(str))
        {
            act_alpha = atof(str);
            console("Set act_alpha %f\n", act_alpha);
        }
    } 



    if (!strcmp("fact", str))                  
    {
        get_next_string(line, index, str);
    
        if (strlen(str))
        {
            fact = atof(str);
            console("Set fact %f\n", fact);
        }
    } 


   
    if (!strcmp("lt_x1", str))                  
    {
        get_next_string(line, index, str);
    
        if (strlen(str))
        {
            lt_x1 = atoi(str);
            console("Set lt_x1 %d\n", lt_x1);
        }
    } 

   
    if (!strcmp("lt_y1", str))                  
    {
        get_next_string(line, index, str);
    
        if (strlen(str))
        {
            lt_y1 = atoi(str);
            console("Set lt_y1 %d\n", lt_y1);
        }
    } 

   
    if (!strcmp("lt_x2", str))                  
    {
        get_next_string(line, index, str);
    
        if (strlen(str))
        {
            lt_x2 = atoi(str);
            console("Set lt_x2 %d\n", lt_x2);
        }
    } 
 
    
    if (!strcmp("lt_y2", str))                  
    {
        get_next_string(line, index, str);
    
        if (strlen(str))
        {
            lt_y2 = atoi(str);
            console("Set lt_y2 %d\n", lt_y2);
        }
    }    
    
    
    
    
    
    
    
    if (!strcmp("rt_x1", str))                  
    {
        get_next_string(line, index, str);
    
        if (strlen(str))
        {
            rt_x1 = atoi(str);
            console("Set rt_x1 %d\n", rt_x1);
        }
    } 

   
    if (!strcmp("rt_y1", str))                  
    {
        get_next_string(line, index, str);
    
        if (strlen(str))
        {
            rt_y1 = atoi(str);
            console("Set rt_y1 %d\n", rt_y1);
        }
    } 

   
    if (!strcmp("rt_x2", str))                  
    {
        get_next_string(line, index, str);
    
        if (strlen(str))
        {
            rt_x2 = atoi(str);
            console("Set rt_x2 %d\n", rt_x2);
        }
    } 
 
    
    if (!strcmp("rt_y2", str))                  
    {
        get_next_string(line, index, str);
    
        if (strlen(str))
        {
            rt_y2 = atoi(str);
            console("Set rt_y2 %d\n", rt_y2);
        }
    }    
    
    
    if (!strcmp("width", str))                  
    {
        get_next_string(line, index, str);
    
        if (strlen(str))
        {
            width = atoi(str);
            console("Set width %d\n", width);
        }
    }    
    
}
/* end of parse_line */


/* loads configuration values from config file */
void load_config(char *config_filename)
{
    FILE *f = fopen(config_filename, "rt");
    
    if (!f)
    {
        add_line("load_config: Unable to load config file\n");
        return;
    }    
    
    console("Loading config file %s\n", config_filename);

    // load each line
    while (!feof(f))
    {
        char line[LINE_MAX];    // temperary line

        unsigned int i = 0;     // index
        
        int line_done = 0;      // line end
        
        // read line
        while (i < LINE_MAX && !line_done)
        {
            char ch = fgetc(f);             // get char from file

            // eat rest of line
            if (ch == '#')
            {
                while (ch != 0x0a && ch != 0x0d && !feof(f))  ch = fgetc(f);
            } else    
    
            // limit check
            if (ch == 0x0d || ch == 0x0a || feof(f)) line_done = 1; else
                line[i++] = ch;
        }            
               
        line[i] = 0; // null terminate
        
        // check for blank line
        if (i > 0)
            parse_line(line);
    }    
    
    fclose(f);
}
/* end of load config */



